;; AUX 1 . CC41A 
;; Oscar E A Callau
;; basado en ejercicios de Richard Ibarra

;; DrScheme -> Language -> Chosee language -> Pretty Big

;; Las líneas iniciadas con ';' son comentarios
;; Puede descomentar las líneas que indican pruebas y ejecutar sus propios experimentos.
;; Para ello puede utilizar la línea de comandos de DrScheme '>'.
;; Asegúrese que entiende la definición de funciones, tras diversas pruebas que ejecute.

;; BOOLEANOS

;; #t -> true
;; #f -> false
;;
;; Funciones lógicas para boleanos : not, and, or

;; LISTAS

;; car : list -> atom          '' Primer elemento de una lista
;; cdr : list -> list          '' Resto de la lista
;; cons : atom x list -> list    '' Retorna una nueva lista añadiendo una elemento al principio
;; append : list x list -> list  '' Retorna una nueva lista resultante de 'pegar' dos listas
;; (list 1 2 3)                '' Crea una nueva lista
;; '(1 2 3)                    '' Igual que el anterior, notar el ' delante del primer paréntesis.
;; '()                         '' Lista vacía
;; (list? l)                   '' Es l una lista?
;; (null? l)                   '' En l una lista vacía?


;; (car '(1 2 3))   -> 1
;; (cdr '(1 2 3))   -> (2 3)
;; (car '())        -> ERROR!
;; (cdr '())        -> ERROR!
;; (car '(1 2))     -> 1
;; (cdr '(1 2))     -> (2)  ... Prop: (number? (cdr '(1 2))) ?  #t | #f ?
;; (car '( (1 2) 3) -> (1 2)
;; (car (cdr '(1 2 3)) -> 2
;; (car (cons 1 '(2 3)) -> 1
;; (append '(1 2) '(3 4)) -> (1 2 3 4)
;; (append 1 '(2 3)) -> ERROR! Ambos argumentos deben ser listas!!!

; -eval :: <poly> x number -> number 
; evalua un polinomio en un punto dado
(define (-eval poly x) (evalExp poly x 0))

; evalExp :: <poly> x number x number -> number
; Evalua un polinomio en un punto dado y con informacion
; de la potencia donde se encuentra.
(define (evalExp poly x exp)
  (if (null? poly) 0
      (let ((coef (car poly)))
        (+ (* coef (pot x exp)) (evalExp (cdr poly) x (+ 1 exp))))))

; pot :: number x number -> number
; calcula la potencia de x elevado al exponente exp
(define (pot x exp)
  (if (eq? exp 0) 1
      (* x (pot x (- exp 1)))))

; -eval* :: <poly> x number -> numbre
; evaluar un polinomio usando foldl
; Problema de esta solución es que solo funciona si los coeficientes son diferentes.
(define (-eval* poly x)
  (foldl (lambda (coef acc) (+ (* coef (pot x (index coef poly 0))) acc))
        0 ; Puedo utilizar (car poly) solo con polinomios validos, no vacios
        poly))

; index :: number  x list -> number
; Busca la posición del elem(number) en la ls(list), dado que este siempre existe.
(define (index elem ls counter)
  (cond 
    ((null? ls) #f) ; nunca se cumple este caso, por pre-condición
    ((eq? elem (car ls)) counter)
    (else (index elem (cdr ls) (+ counter 1)))))

; -eval** :: <poly> x number -> numbre
; evaluar un polinomio usando foldl
; Solución elegante de -eval*, funciona en todos los casos.
(define (-eval** poly x)
   (foldl (lambda (coef acc) (+ (* (first coef) ; Si no lo reconoce usar (car coef) 
                                   (pot x 
                                        (second coef))) ; Si no lo reconoce usar (car (cdr coef)) 
                               acc))
         0
         (ls-pair poly 0)))

;ls-pair :: list x number -> list-pair
(define (ls-pair ls counter)
  (if (null? ls) 
      '()
      (cons (list (car ls) counter) (ls-pair (cdr ls) (+ counter 1)))))


; -nlist :: number x number -> list
; retorna la lista que incluye todos los enteros entre dos números
(define (-nlist n1 n2)
  (cond
    ((> n1 n2) '())
    (else (cons n1 (-nlist (+ n1 1) n2)))
    )
  )

;; (-nlist 1 5)

; -reverse :: list -> list
(define (-reverse ls)
  (if (null? ls) '() (append (-reverse (cdr ls)) (list (car ls)))))

;; my-member :: atom x list -> boolean
;; pregunta si un elemento pertenece o no a una lista.
(define (my-member atom list)
  (cond
    ((null? list) #f)
    ((eq? (car list) atom) #t)
    (else (my-member atom (cdr list)))
    ))


(my-member 1 '(1 2 3))
(my-member 5 '(1 2 3))
(my-member 5 '(1 2 3 (5)))

;; eq? : Permite determinar si dos objetos en Scheme son el mismo.

; my-member-list :: atom x list -> boolean 
; Soluciona el problema de my-member en listas anidadas
(define (my-member-list atom list)
  (cond
    ((null? list) #f)
    ((list? (car list))
     (or 
      (my-member-list atom (car list)) ;; Permite 'entrar' en las sublistas
      (my-member-list atom (cdr list))
      ))
    ((eq? (car list) atom) #t)
    (else (my-member-list atom (cdr list)))
    )
  )


(my-member-list 5 '(1 2 3 (5)))
(my-member-list 5 '(1 (5 6) 2 ))
(my-member-list 5 '(1 2 (3 4)))

; my-member-list* :: atom x list -> boolean 
; Soluciona el problema de my-member en listas anidadas, usan let
(define (my-member-list* atom list)
  (let ((head (lambda () (car list)))
        (tail (lambda () (cdr list))))
  (cond
    ((null? list) #f)
    ((list? (head))
     (or 
      (my-member-list* atom (head)) ;; Permite 'entrar' en las sublistas
      (my-member-list* atom (tail))))
    ((eq? (head) atom) #t)
    (else (my-member-list* atom (tail))))))


(my-member-list* 5 '(1 2 3 (5)))
(my-member-list* 5 '(1 (5 6) 2 ))
(my-member-list* 5 '(1 2 (3 4)))